#pragma once

#include <amdis/GridFunctions.hpp>
#include <amdis/gridfunctions/DiscreteFunction.hpp>

namespace AMDiS
{
  /// A mutable view on the subspace of a DOFVector, \relates DiscreteFunction
  template <class GlobalBasisType, class RangeType, class TreePathType>
  class DOFVectorView
      : public DiscreteFunction<GlobalBasisType, RangeType, TreePathType>
  {
    using Self = DOFVectorView;
    using Super = DiscreteFunction<GlobalBasisType, RangeType, TreePathType>;

    using GlobalBasis = GlobalBasisType;
    using TreePath = TreePathType;

  public:
    /// Constructor. Stores a pointer to the mutable `dofvector`.
    DOFVectorView(DOFVector<GlobalBasis,RangeType>& dofVector, TreePath const& treePath)
      : Super(dofVector, treePath)
      , mutableDofVector_(&dofVector)
    {}

  public:
    /// \brief Interpolation of GridFunction to DOFVector, assuming that there is no
    /// reference to this DOFVector in the expression.
    template <class Expr>
    void interpolate_noalias(Expr&& expr)
    {
      auto const& basis = this->basis();
      auto const& treePath = this->treePath();

      auto&& gridFct = makeGridFunction(std::forward<Expr>(expr), basis.gridView());

      Dune::Functions::interpolate(basis, treePath, coefficients(), std::forward<decltype(gridFct)>(gridFct));
    }

    /// Interpolation of GridFunction to DOFVector
    template <class Expr>
    void interpolate(Expr&& expr)
    {
      // create temporary copy of data
      DOFVector<GlobalBasis,RangeType> tmp(coefficients());
      Self tmpView{tmp, this->treePath()};
      tmpView.interpolate_noalias(std::forward<Expr>(expr));

      // move data from temporary vector into stored DOFVector
      coefficients().vector() = std::move(tmp.vector());
    }

    /// \brief Interpolation of GridFunction to DOFVector, alias to \ref interpolate()
    template <class Expr>
    DOFVectorView& operator<<(Expr&& expr)
    {
      interpolate(expr);
      return *this;
    }

    template <class Expr>
    DOFVectorView& operator+=(Expr&& expr)
    {
      interpolate((*this) + expr);
      return *this;
    }

    template <class Expr>
    DOFVectorView& operator-=(Expr&& expr)
    {
      interpolate((*this) - expr);
      return *this;
    }


    /// Return the mutable DOFVector
    DOFVector<GlobalBasis,RangeType>& coefficients() { return *mutableDofVector_; }

    /// Return the const DOFVector
    using Super::coefficients;

  protected:
    DOFVector<GlobalBasis,RangeType>* mutableDofVector_;
  };


  /// A Generator for a mutable \ref DOFVectorView
  template <class GlobalBasis, class RangeType, class TreePath>
  auto makeDOFVectorView(DOFVector<GlobalBasis, RangeType>& dofVector, TreePath const& treePath)
  {
    return DOFVectorView<GlobalBasis, RangeType, TreePath>{dofVector, treePath};
  }

  /// A Generator for a mutable \ref DOFVectorView
  template <class GlobalBasis, class RangeType>
  auto makeDOFVectorView(DOFVector<GlobalBasis, RangeType>& dofVector)
  {
    auto treePath = Dune::TypeTree::hybridTreePath();
    return DOFVectorView<GlobalBasis, RangeType, decltype(treePath)>{dofVector, treePath};
  }

} // end namespace AMDiS
