#pragma once

#include <algorithm>
#include <string>

#include <amdis/Output.hpp>
#include <amdis/linear_algebra/RunnerInterface.hpp>
#include <amdis/linear_algebra/SolverInfo.hpp>

namespace AMDiS
{
  /**
   * \ingroup Solver
   * \class AMDiS::DirectRunner
   * \brief \implements RunnerInterface for the (external) direct solvers
   */
  template <class Solver, class Matrix, class VectorX, class VectorB>
  class DirectRunner
      : public RunnerInterface<Matrix, VectorX, VectorB>
  {
  protected:
    using Super = RunnerInterface<Matrix, VectorX, VectorB>;
    using BaseSolver  = Dune::InverseOperator<VectorX, VectorB>;

  public:
    /// Constructor.
    DirectRunner(std::string prefix)
      : solverCreator_(prefix)
    {}

    /// Implementes \ref RunnerInterface::init()
    virtual void init(Matrix const& A) override
    {
      solver_ = solverCreator_.create(A);
    }

    /// Implementes \ref RunnerInterface::exit()
    virtual void exit() override
    {
      solver_.reset();
    }

    /// Implementes \ref RunnerInterface::solve()
    virtual int solve(Matrix const& A, VectorX& x, VectorB const& b,
                      SolverInfo& solverInfo) override
    {
      // storing some statistics
      Dune::InverseOperatorResult statistics;

      // solve the linear system
      VectorB _b = b;
      solver_->apply(x, _b, statistics);

      solverInfo.setRelResidual(statistics.reduction);

      return 0;
    }

  private:
    ISTLSolverCreator<Solver> solverCreator_;
    std::shared_ptr<BaseSolver> solver_;
  };
}
